// CyRC v1.11 Cyrillic CodePage Recoder
// Copyright (c) 1998 SeaD

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "service.h"

#define NAME_STRING      "CyRC v1.11 FREEWARE Cyrillic CodePage Recoder\n"
#define COPYRIGHT_STRING "Copyright (c) 1998 SeaD\n"
#define USAGE_STRING     "Usage: cyrc [/<src_cp> [/<dst_cp>]] <src_file> [<dst_file>]\n"
#define HELP_STRING      "Help: cyrc /?, cyrc -h, cyrc /help\n"
#define KEYWORDS_STRING  "Standart keywords:\n"
#define TRANSKEYS_STRING "Translit ..................................... 0,T,TRANS,TRANSLIT\n"
#define DOSKEYS_STRING   "Dos 866 codepage ............. 1,D,DOS,DOS866,866,CP866,ALT,O,OS2\n"
#define WINKEYS_STRING   "Windows 1251 codepage ....... 2,W,WIN,WINDOWS,WIN1251,1251,CP1251\n"
#define UNIXKEYS_STRING  "Unix Koi8 codepage ........................ 3,K,KOI8,KOI8R,U,UNIX\n"
#define HELPKEYS_STRING  "This Help Screen ....................................... ?,H,HELP\n"
#define EXAMPLES_STRING1 "Examples: cyrc -k -w koi_text.txt win_text.txt\n"
#define EXAMPLES_STRING2 "          cyrc win_html.htm\n"
#define FORHELP_STRING   "For more help see cyrc.txt\n"

#define DEFAULT_NAME   "default"
#define DEFAULT_SRC_CP "sourcecodepage"
#define DEFAULT_DST_CP "destinationcodepage"
#define CODEPAGE_ID    "codepageid"
#define LOWERCASE_CHR  "lowercasechar"
#define UPPERCASE_CHR  "uppercasechar"
#define LOWERCASE_INT  "lowercaseint"
#define UPPERCASE_INT  "uppercaseint"

#define INISTRING_MAXLEN 256
#define PARAM_MAXLEN     16
#define CODEPAGES_MAXNUM 64
#define OPTION_MAXLEN    32
#define NAME_MAXLEN      32
#define ID_MAXNUM        16
#define ID_MAXLEN        16
#define LETTER_MAXNUM    64
#define LETTER_MAXLEN    8

typedef struct letters_register_struct {
  int str_num;
  char letter[LETTER_MAXNUM][LETTER_MAXLEN];
  } let_typ, *let_ptr;

typedef struct cyrillic_codepage_struct {
  char name[NAME_MAXLEN];
  char id[ID_MAXNUM][ID_MAXLEN];
  let_typ lowercase;
  let_typ uppercase;
  } cyr_typ, *cyr_ptr;

cyr_ptr cp_src = NULL, cp_dst = NULL;
FILE *ini_fil;
char name_src[13], name_dst[13];
char param_src[PARAM_MAXLEN] = "\0", param_dst[PARAM_MAXLEN] = "\0";

void Tip_Screen(void) {
  printf(USAGE_STRING);
  printf(HELP_STRING);
  exit(EXIT_SUCCESS);
  }

void Help_Screen(void) {
  printf(USAGE_STRING);
  printf("\n");
  printf(KEYWORDS_STRING);
  printf(TRANSKEYS_STRING);
  printf(DOSKEYS_STRING);
  printf(WINKEYS_STRING);
  printf(UNIXKEYS_STRING);
  printf(HELPKEYS_STRING);
  printf("\n");
  printf(EXAMPLES_STRING1);
  printf(EXAMPLES_STRING2);
  printf("\n");
  printf(FORHELP_STRING);
  exit(EXIT_SUCCESS);
  }

void Error_Exit(char *error_str) {
  printf("Error: %s.\n", error_str);
  exit(EXIT_FAILURE);
  }

void Create_IniFile(char *str) {
  if (!(ini_fil = fopen(str, "w+t")))
    Error_Exit("can't create .ini file");
  fprintf(ini_fil, "; CyRC v1.11 Data File\n\n[default]\n; Default Source CodePage Identifier\nSourceCodePage=cp1251\n; Default Destination CodePage Identifier\nDestinationCodePage=cp866\n\n");
  fprintf(ini_fil, "; CodePage Name\n; CodePage Identifiers\n; \n; \n\n");
  fprintf(ini_fil, "; Transliteration\n[Translit]\nCodePageId=0,t,trans,translit\n");
  printf(ini_fil, "LowerCaseChar=a,b,v,g,d,e,zh,z,i,y,k,l,m,n,o,p,r,s,t,u,f,kh,ts,ch,sh,tsh,,y,,e,yu,ya,e\n");
  fprintf(ini_fil, "UpperCaseChar=A,B,V,G,D,E,ZH,Z,I,Y,K,L,M,N,O,P,R,S,T,U,F,KH,TS,CH,SH,TSH,,Y,,E,YU,YA,E\n\n");
  fprintf(ini_fil, "; Dos, Os2 866 CodePage\n[Dos,Os/2 866]\nCodePageId=1,d,dos,dos866,866,cp866,alt,o,os2\n");
  fprintf(ini_fil, "LowerCaseChar=,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,\n");
  fprintf(ini_fil, "UpperCaseChar=,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,\n\n");
  fprintf(ini_fil, "; Windows 1251 CodePage\n[Windows 1251]\nCodePageId=2,w,win,windows,win1251,1251,cp1251\n");
  fprintf(ini_fil, "LowerCaseInt=224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255,184\n");
  fprintf(ini_fil, "UpperCaseInt=192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,168\n\n");
  fprintf(ini_fil, "; Unix koi8 CodePage\n[Unix Koi-8]\nCodePageId=3,k,koi8,koi8r,u,unix\n");
  fprintf(ini_fil, "LowerCaseInt=193,194,215,199,196,197,214,218,201,202,203,204,205,206,207,208,210,211,212,213,198,200,195,222,219,221,223,217,216,220,192,209,163\n");
  fprintf(ini_fil, "UpperCaseInt=225,226,247,231,228,229,246,250,233,234,235,236,237,238,239,240,242,243,244,245,230,232,227,254,251,253,255,249,248,252,224,241,179\n\n");
  fseek(ini_fil, 0l, SEEK_SET);
  }

char Get_Char(void) {
  char str[INISTRING_MAXLEN], chr;
  while ((chr = fgetc(ini_fil))!=EOF) {
    if (chr==' ') continue;
    if (chr=='\n') continue;
    if (chr==';') {
      fgets(str, INISTRING_MAXLEN, ini_fil); continue;
      }
    return (chr);
    }
  return (0);
  }

char *Get_Name(char *str) {
  int n;
  if ((str[0] = Get_Char())=='[')
    for (n = 0; n<NAME_MAXLEN; n++) {
      if ((str[n] = fgetc(ini_fil))==']') {
        str[n] = '\0'; return (str);
        }
      }
  Error_Exit("wrong .ini file");
  }

void Get_Option(char *str) {
  int n;
  char chr;
  str[0] = Get_Char();
  for (n = 1; n<OPTION_MAXLEN;) {
    if ((chr = fgetc(ini_fil))==' ') continue;
    if (chr=='=') {
      str[n] = '\0'; return (strlwr(str));
      }
    str[n++] = chr;
    }
  Error_Exit("wrong .ini file");
  }

void Get_Default_Param(char *str) {
  int n;
  char chr;
  for (n = 0; n<PARAM_MAXLEN;) {
    if ((chr = fgetc(ini_fil))==' ') continue;
    if (chr=='\n') {
      str[n] = '\0'; return (strlwr(str));
      }
    str[n++] = chr;
    }
  Error_Exit("wrong .ini file");
  }

void Default_Params(void) {
  char str[OPTION_MAXLEN];
  if (strcmp(strlwr(Get_Name(str)), DEFAULT_NAME))
    Error_Exit("wrong .ini file");
  while (param_src[0]=='\0' || param_dst[0]=='\0') {
    Get_Option(str);
    if (!strcmp(str, DEFAULT_SRC_CP)) {
      Get_Default_Param(param_src); continue;
      }
    if (!strcmp(str, DEFAULT_DST_CP)) {
      Get_Default_Param(param_dst); continue;
      }
    Error_Exit("wrong .ini file");
    }
  }

void Cmd_Params(int args, char *argv[]) {
  char *str = "\0";
  if (!--args)
    Tip_Screen();
  strcpy(str, strlwr(argv[1]));
  if (str[0]=='/' || str[0]=='-')
    strcpy(param_src, ++str);
  else {
    strcpy(name_src, str);
    if (!--args) {
      strcpy(name_dst, name_src); return;
      }
    strcpy(name_dst, argv[2]);
    if (!--args) return;
    Error_Exit("invalid parameters");
    }
  if (!--args) {
    if (!strcmp(str, "?") ||
        !strcmp(str, "h") ||
        !strcmp(str, "help"))
      Help_Screen();
    else
      Error_Exit("invalid parameters");
    }
  strcpy(str, strlwr(argv[2]));
  if (str[0]=='/' || str[0]=='-')
    strcpy(param_dst, ++str);
  else {
    strcpy(name_src, str);
    if (!--args) {
      strcpy(name_dst, name_src); return;
      }
    strcpy(name_dst, argv[3]);
    if (!--args) return;
    Error_Exit("invalid parameters");
    }
  if (!--args)
    Error_Exit("invalid parameters");
  strcpy(name_src, argv[3]);
  if (!--args) {
    strcpy(name_dst, name_src); return;
    }
  strcpy(name_dst, argv[4]);
  if (!--args) return;
  Error_Exit("invalid parameters");
  }

char *Get_Param(char *str) {
  int n;
  char chr;
  for (n = 0; n<PARAM_MAXLEN;) {
    if ((chr = fgetc(ini_fil))==' ') continue;
    if (chr==',') {
      str[n] = '\0'; return (str);
      }
    if (chr=='\n') {
      fseek(ini_fil, -1, SEEK_CUR);
      str[n] = '\0'; return (str);
      }
    str[n++] = chr;
    }
  Error_Exit("wrong .ini file");
  }

int CodePage_Sort(let_ptr let_src, let_ptr let_dst) {
  int n, len, num;
  char str[LETTER_MAXLEN];
  for (len = LETTER_MAXLEN - 1, num = 0; len!=1; len--)
    for (n = 0; n<LETTER_MAXNUM; n++)
      if (strlen(let_src->letter[n])==len) {
        strcpy(str, let_src->letter[n]);
        strcpy(let_src->letter[n], let_src->letter[num]);
        strcpy(let_src->letter[num], str);
        strcpy(str, let_dst->letter[n]);
        strcpy(let_dst->letter[n], let_dst->letter[num]);
        strcpy(let_dst->letter[num++], str);
        }
  return (num);
  }

void Get_CodePages(cyr_ptr cp1, cyr_ptr cp2) {
  cyr_ptr cp = cp1;
  int n, count;
  int id, lowercase, uppercase;
  char str[OPTION_MAXLEN];
  for (count = 0; count<LETTER_MAXNUM; count++) {
    Get_Name(cp->name);
    id = lowercase = uppercase = TRUE;
    while (id || lowercase || uppercase) {
      if (feof(ini_fil))
        Error_Exit("wrong .ini file");
      Get_Option(str);
      if (!strcmp(str, CODEPAGE_ID)) {
        for (n = 0; n<ID_MAXNUM; Get_Param(cp->id[n++]));
        id = FALSE; continue;
        }
      if (!strcmp(str, LOWERCASE_CHR)) {
        for (n = 0; n<LETTER_MAXNUM; Get_Param(cp->lowercase.letter[n++]));
        cp->lowercase.str_num = TRUE;
        lowercase = FALSE; continue;
        }
      if (!strcmp(str, UPPERCASE_CHR)) {
        for (n = 0; n<LETTER_MAXNUM; Get_Param(cp->uppercase.letter[n++]));
        cp->uppercase.str_num = TRUE;
        uppercase = FALSE; continue;
        }
      if (!strcmp(str, LOWERCASE_INT)) {
        for (n = 0; n<LETTER_MAXNUM; cp->lowercase.letter[n][0] = atoi(Get_Param(str)),
                                     cp->lowercase.letter[n++][1] = '\0');
        lowercase = cp->lowercase.str_num = FALSE; continue;
        }
      if (!strcmp(str, UPPERCASE_INT)) {
        for (n = 0; n<LETTER_MAXNUM; cp->uppercase.letter[n][0] = atoi(Get_Param(str)),
                                     cp->uppercase.letter[n++][1] = '\0');
        uppercase = cp->uppercase.str_num = FALSE; continue;
        }
      Error_Exit("wrong .ini file");
      }
    for (n = 0; n<ID_MAXNUM; n++) {
      if (!strcmp(param_src, cp->id[n])) cp_src = cp;
      if (!strcmp(param_dst, cp->id[n])) cp_dst = cp;
      }
    if (cp_src && cp_dst) break;
    if (cp_src || cp_dst) cp = cp2;
    }
  if (!cp_src || !cp_dst)
    Error_Exit("invalid parameters");
  cp_src->lowercase.str_num = CodePage_Sort(&cp_src->lowercase, &cp_dst->lowercase);
  cp_src->uppercase.str_num = CodePage_Sort(&cp_src->uppercase, &cp_dst->uppercase);
  }

void Cyr_Recode(void) {
  FILE *src_fil, *dst_fil;
  int n, len;
  unsigned long count, count_max, pos;
  char *str = "wb";
  if (!strcmp(name_src, name_dst)) strcpy(str, "r+b");
  if ((src_fil = fopen(name_src, "rb"))==NULL)
    Error_Exit("can't open source file");
  if ((dst_fil = fopen(name_dst, str))==NULL)
    Error_Exit("can't open destination file");
  fseek(src_fil, 0, SEEK_END);
  count = (count_max = ftell(src_fil) / 32) / 2;
  fseek(src_fil, 0, SEEK_SET);
  printf("Source file:       %s (%s codepage)\n", name_src, cp_src->name);
  printf("Destination file:  %s (%s codepage)\n", name_dst, cp_dst->name);
  printf("\n                                                    ]\r");
  printf("Recoding progress: [");
  if (cp_src->lowercase.str_num || cp_src->uppercase.str_num)
    while(TRUE) {
      for (len = LETTER_MAXLEN - 1, pos = ftell(src_fil), fgets(str, len + 1, src_fil); ; len--) {
        for (n = 0; n<LETTER_MAXNUM; n++)
          if (!strcmp(str, cp_src->lowercase.letter[n])) {
            strcpy(str, cp_dst->lowercase.letter[n]); break;
            }
        if (n!=LETTER_MAXNUM) break;
        for (n = 0; n<LETTER_MAXNUM; n++)
          if (!strcmp(str, cp_src->uppercase.letter[n])) {
            strcpy(str, cp_dst->uppercase.letter[n]); break;
            }
        if (n!=LETTER_MAXNUM) break;
        if (len==1) break;
        str[len-1] = '\0';
        }
      if (feof(src_fil)) break;
      fseek(src_fil, pos+len, SEEK_SET);
      fputs(str, dst_fil);
      if ((count += len)>=count_max) {
        printf("."); count = 0;
        }
      }
  else
    while(TRUE) {
      str[1] = '\0';
      if ((str[0] = fgetc(src_fil))<0)
        for (n = 0; n<LETTER_MAXNUM; n++) {
          if (str[0]==cp_src->lowercase.letter[n][0]) {
            strcpy(str, cp_dst->lowercase.letter[n]); break;
            }
          if (str[0]==cp_src->uppercase.letter[n][0]) {
            strcpy(str, cp_dst->uppercase.letter[n]); break;
            }
          }
      if (feof(src_fil)) break;
      fputs(str, dst_fil);
      if (count++==count_max) {
        printf("."); count = 0;
        }
      }
  fclose(dst_fil);
  fclose(src_fil);
  }

int main(int args, char *argv[]) {
  cyr_typ cp1, cp2;
  char str[80];
  printf(NAME_STRING);
  printf(COPYRIGHT_STRING);
  printf("\n");
  if (!strcpy(strstr(strrchr(strcpy(str, strlwr(argv[0])), '\\'), "exe"), "ini"))
    Error_Exit("can't find .exe file");
  if (!(ini_fil = fopen(str, "rt")))
    Create_IniFile(str);
  Default_Params();
  Cmd_Params(args, argv);
  Get_CodePages(&cp1, &cp2);
  fclose(ini_fil);
  Cyr_Recode();
  printf("\nOK.\n");
  return(0);
  }
